#include <Foundation/Foundation.h>
#include <AppKit/AppKit.h>
#include "AppController.h"
#include "FileRing.h"
#include "ImageWindow.h"
#include "SelectionCtrl.h"
#include "PHDefs.h"

@interface AppController (PrivateMethods)

- (void) setCheckMarkInMenu: (NSMenu *) menu forValue: (float) value;
- (void) addToMenu: (NSMenu *) menu fromDescr: (NSString *[][3]) descr
            amount: (int) amount selector: (SEL) sel;

@end // @interface AppController (PrivateMethods)

@implementation AppController

- (id) init
{
    [super init];

    _window = nil;
    _fileRing = nil;
    _selCtrl = [SelectionCtrl new];

    return self;
}

- (void) dealloc
{
    RELEASE(_selCtrl);
    RELEASE(_fileRing);
    RELEASE(_window);
}

// {title, key equivalent, shape (y/x ratio)}
static NSString * shape_data[][3] =
{
    {@"No constraint", @"", @"0"},
    {@"Monitor (4:3)", @"", @"0.75"},
    {@"Wide (16:9)", @"", @"0.5625"},
    {@"Square (1:1)", @"", @"1"}
};

// {title, key equivalent, max. size after scale}
static NSString * scale_data[][3] =
{
    {@"Do not scale", @"", @"0"},
    //{@"Maximal standard size", @"", @"-1"},
    {@"2272", @"", @"2272"},
    {@"2048", @"", @"2048"},
    {@"1600", @"", @"1600"},
    {@"1280", @"", @"1280"},
    {@"1024", @"", @"1024"},
    {@"800", @"", @"800"},
    {@"640", @"", @"640"},
    {@"480", @"", @"480"},
    {@"320", @"", @"320"},
};

- (void) createMenu
{
    SEL dummy = @selector(dummy_method:);
    
    NSMenu * menu = AUTORELEASE( [NSMenu new] );

    [menu addItemWithTitle: @"Info Panel..."
          action: @selector(showInfoPanel:)
          keyEquivalent: @""];
    [menu addItemWithTitle: @"Selection"
          action: dummy
          keyEquivalent: @""];
    [menu addItemWithTitle: @"Open"
          action: @selector(open:)
          keyEquivalent: @"o"];
    [menu addItemWithTitle: @"Rotate"
          action: @selector(rotatePositive:)
          keyEquivalent: @"r"];
    [menu addItemWithTitle: @"Rotate back"
          action: @selector(rotateNegative:)
          keyEquivalent: @"R"];
    [menu addItemWithTitle: @"Next"
          action: @selector(nextImage:)
          keyEquivalent: @"n"];
    [menu addItemWithTitle: @"Previous"
          action: @selector(prevImage:)
          keyEquivalent: @"p"];
    [menu addItemWithTitle: @"Clip"
          action: @selector(clip:)
          keyEquivalent: @"c"];
    [menu addItemWithTitle: @"Undo"
          action: @selector(undo:)
          keyEquivalent: @"z"];
    [menu addItemWithTitle: @"Save"
           action: @selector(save:)
           keyEquivalent: @"s"];
    [menu addItemWithTitle: @"Hide"
          action: @selector(hide:)
          keyEquivalent: @"h"];
    [menu addItemWithTitle:@"Quit"
          action:@selector(terminate:)
          keyEquivalent:@"q"];


    // "Shape" menu
    
    NSMenu * shape = AUTORELEASE( [NSMenu new] );

    // Add shapes
    
    int num_shapes = sizeof(shape_data)/sizeof(*shape_data);
    [self addToMenu: shape fromDescr: shape_data amount: num_shapes
          selector: @selector(setSelectionShape:)];

    // Add more items to shape menu "manually"
    [shape addItemWithTitle: @"Scale To" action: dummy keyEquivalent: @""];

    [self setCheckMarkInMenu: shape forValue: [_selCtrl shape] ];

    // Attach submenu 
    [menu setSubmenu: shape forItem: [menu itemWithTitle:@"Selection"]];
    
    // "Scale To" menu
    
    NSMenu * scale_to = AUTORELEASE( [NSMenu new] );

    int num_scales = sizeof(scale_data)/sizeof(*scale_data);
    [self addToMenu: scale_to fromDescr: scale_data amount: num_scales
          selector: @selector(setSelectionScaleTo:)];

    [self setCheckMarkInMenu: scale_to forValue: [_selCtrl bound] ];

    // Comment ScaleTo for cleaner release while no functionality
    // Attach submenu
    // [menu setSubmenu: scale_to forItem: [shape itemWithTitle:@"Scale To"]];
    
    [NSApp setMainMenu: menu];
}

- (void) addToMenu: (NSMenu *) menu fromDescr: (NSString *[][3]) descr
            amount: (int) amount selector: (SEL) sel
{
    NSMenuItem * item;
    int i;
    for (i=0; i < amount; ++i)
    {
        item = [[NSMenuItem alloc] initWithTitle: descr[i][0]
                               action: sel
                               keyEquivalent: descr[i][1]];
        float value = [descr[i][2] floatValue];
        [item setRepresentedObject: [NSNumber numberWithFloat: value]];

        [menu addItem: item];
        RELEASE(item);
    }
}

- (void) createWindow
{
    _window = [[ImageWindow alloc] init];
}

- (void) applicationWillFinishLaunching: (NSNotification *)not
{
    [self createMenu];
    [self createWindow];

    // get first command-line argument (zeroth is the program name)
    NSArray * args = [[NSProcessInfo processInfo] arguments];
    if ([args count] > 1)
        [self openFile: [args objectAtIndex:1] ];
}

- (void) applicationDidFinishLaunching: (NSNotification *)not
{
    [_window center];
    [_window makeKeyAndOrderFront: nil];
}

- (void) applicationWillTerminate: (NSNotification *)not
{
    // TODO: remove after making proper cursor handling.
    
    // A workaround for cursor handling bug. Reset to default.
    [[NSCursor arrowCursor] set];
}


- (BOOL) applicationShouldTerminateAfterLastWindowClosed: (id)sender
{
    return YES;
}

- (void) open: (id)sender
{
    NSArray * image_types = [NSImage imageFileTypes];
    NSOpenPanel * panel = [NSOpenPanel openPanel];
    [panel setCanChooseDirectories: NO];

    NSString * dir = (_fileRing) ? [_fileRing directory] : NSHomeDirectory();

    NSString * fileName = (_fileRing) ? [_fileRing currentName] : nil;

    int result = [panel runModalForDirectory: dir
                        file: fileName
                        types: image_types ];
    if (result == NSOKButton)
    {
        NSArray * selection = [panel filenames];
        NSString * sel_file = [selection objectAtIndex: 0];
        if (sel_file)
            [self openFile: sel_file];
    }
}

- (void) openFile: (NSString *)fileName
{
    [self openFile: fileName updateRing: YES];
}

- (void) openFile: (NSString *)fileName updateRing: (BOOL) update
{
    NSAssert(_window, @"open: '_window' is not initialized");

    if (update)
    {
        RELEASE(_fileRing);
        _fileRing = [[FileRing alloc] initFromPath: fileName
                                      withTypes: [NSImage imageFileTypes]];
    }
    
    [_window setImageFromFile: fileName];
}

- (void) rotatePositive: (id)sender
{
    NSAssert(_window, @"rotatePositive: '_window' is not initialized");

    [_window rotateWithDirection: PHPositiveDirection];
}

- (void) rotateNegative: (id)sender
{
    NSAssert(_window, @"rotateNegative: '_window' is not initialized");

    [_window rotateWithDirection: PHNegativeDirection];
}

- (void) nextImage: (id)sender
{
    [self openFile: [_fileRing nextFile] updateRing: NO];
}

- (void) prevImage: (id)sender
{
    [self openFile: [_fileRing prevFile] updateRing: NO];
}

- (void) clip: (id)sender
{
    NSAssert(_window, @"clip: '_window' is not initialized");

    [_window clip];
}

- (void) undo: (id)sender
{
    NSAssert(_window, @"undo: '_window' is not initialized");

    [_window undo];
}

- (void) save: (id)sender
{
    NSAssert(_window, @"save: '_window' is not initialized");
    [_window saveToFile: @"tmpsave.tiff"];
}

- (void) setSelectionShape: (id)sender
{
    float shape = 0;
    id rep = [sender representedObject];
    if (rep)
        shape = [rep floatValue];

    [_selCtrl setShape: shape];

    [self setCheckMarkInMenu: [sender menu] forValue: shape];
}

- (void) setSelectionScaleTo: (id)sender
{
    float bound = 0;
    id rep = [sender representedObject];
    if (rep)
        bound = [rep floatValue];

    [_selCtrl setBound: bound];

    [self setCheckMarkInMenu: [sender menu] forValue: bound];
}

#define IS_CLOSE(a, b, tol) (fabs((a) - (b)) < tol)

- (void) setCheckMarkInMenu: (NSMenu *) menu forValue: (float) value
{
    Class number_class = [NSNumber class];

    NSEnumerator * items = [[menu itemArray] objectEnumerator];
    NSMenuItem * item;

    float current_shape = [_selCtrl shape];
    const float TOL = 0.01;

    while (item = [items nextObject])
    {
        id rep_object = [item representedObject];
        if ([rep_object isKindOfClass: number_class])
        {
            int state = IS_CLOSE([rep_object floatValue], value, TOL) ?
                NSOnState : NSOffState;

            [item setState: state];
        }
    }
}

- (void) showInfoPanel: (id)sender
{
    NSString * image_types = [[NSImage imageFileTypes] description];
    NSString * app_descr =
        [NSString stringWithFormat:
                      @"Simple image viewer. Reads %@", image_types];

    NSDictionary * options =
        [NSDictionary dictionaryWithObjectsAndKeys:
                      @"Simple image viewer", @"ApplicationDescription",
                      @"http://www.vaisburd.net/PhotoClip/", @"URL",
                      [NSArray arrayWithObject:
                                   @"Tima Vaisburd <tima@vaisburd.net>"],
                      @"Authors",
                      @"0.2.2", @"ApplicationVersion",
                      @"Copyright (C) 2003 Tima Vaisburd", @"Copyright",
                      @"License: GNU GPL", @"CopyrightDescription",
                      nil];
                      
                                               
    [NSApp orderFrontStandardInfoPanelWithOptions: options];
}

- (SelectionCtrl *) selectionCtrl
{
    return _selCtrl;
}

@end
